// Copyright 2002-2012 by Jon Dart. All Rights Reserved.
#include "options.h"

#include <fstream>
#include <ctype.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <iterator>

using namespace std;

const string Options::NALIMOV_TYPE = "Nalimov";
const string Options::GAVIOTA_TYPE = "Gaviota";

static void set_int_option(const string &name,
			   int &value, const string &valueString) {
  if (sscanf(valueString.c_str(),"%d",&value) != 1)
    cerr << "warning: invalid value for option " << name << " (expected integer)"  << endl;
}

static void set_strength_option(const string &name,
			   int &value, const string &valueString) {
  int tmp = value;
  if (sscanf(valueString.c_str(),"%d",&tmp) != 1 || tmp < 0 || tmp > 100) {
    cerr << "warning: invalid value for option " << name << " (expected integer 0..100)"  << endl;
    return;
  }
  value = tmp;
}

void Options::setMemoryOption(size_t &value, const string &valueString) {
  string s = valueString;
  string::iterator it = s.end()-1;
  size_t mult = 1L;
  if (*it == 'k' || *it == 'K') {
    s.erase(it);
    mult = 1024L;
  } 
  else if (*it == 'm' || *it == 'M') {
    s.erase(it);
    mult = 1024L*1024L;
  }
  else if (*it == 'g' || *it == 'G') {
    s.erase(it);
    mult = 1024L*1024L*1024L;
  }
  stringstream ss(s);
  size_t val;
  if((ss >> val).fail())
     return; // invalid value
  else
     value = (size_t)val*mult;
}

static void set_boolean_option(string name,int &value, string valueString) {
  if (valueString == "true")
    value = 1;
  else if (valueString == "false") 
    value = 0;
  else
    cerr << "warning: invalid value for option " << name << " (expected 'true' or 'false')"  << endl;
}

void Options::set_option(const string &name, const string &value) {
  if (name == "log.enabled") {
    set_boolean_option(name,log_enabled,value);
  }
  else if (name == "log.append") {
    set_boolean_option(name,log_append,value);
  }
  else if (name == "log.pathName") {
    log_pathname = value;
  }
  else if (name == "store_games") {
    set_boolean_option(name,store_games,value);
  }
  else if (name == "game_pathname") {
    game_pathname = value;
  }
  else if (name == "book.book_enabled") {
    set_boolean_option(name,book.book_enabled,value);
  }
  else if (name == "book.selectivity") {
    set_int_option(name,book.selectivity,value);
  }
  else if (name == "book.random") {
    set_int_option(name,book.random,value);
  }
  else if (name == "learning.score_learning") {
    set_boolean_option(name,learning.score_learning,value);
  }
  else if (name == "learning.score_learning.horizon") {
    set_int_option(name,learning.score_learning_horizon,value);
  }
  else if (name == "learning.score_learning.threshold") {
    int tmp;
    set_int_option(name,tmp,value);
    learning.score_learning_threshold = (64*tmp)/100;
  }
  else if (name == "learning.score_learning.minDepth") {
    set_int_option(name,learning.score_learning_minDepth,value);
  }
  else if (name == "learning.result_learning") {
    set_boolean_option(name,learning.result_learning,value);
  }
  else if (name == "learning.position_learning") {
    set_boolean_option(name,learning.position_learning,value);
  }
  else if (name == "learning.position_learning.threshold") {
    int tmp;
    set_int_option(name,tmp,value);
    learning.position_learning_threshold = (64*tmp)/100;
  }
  else if (name == "learning.position_learning.minDepth") {
    set_int_option(name,learning.position_learning_minDepth,value);
  }
  else if (name == "search.checks_in_qsearch") {
    set_int_option(name,search.checks_in_qsearch,value);
  }
  else if (name == "search.can_resign") {
    set_boolean_option(name,search.can_resign,value);
  }
  else if (name == "search.resign_threshold") {
    set_int_option(name,search.resign_threshold,value);
  }
  else if (name == "search.hash_table_size") {
    setMemoryOption(search.hash_table_size,value);
  }
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
  else if (name == "search.use_tablebases") {
    set_boolean_option(name,search.use_tablebases,value);
  }
  else if (name == "search.tablebase_type") {
      string tmp = search.tablebase_type;
      search.tablebase_type.clear();
#ifdef GAVIOTA_TBS
      if (value == GAVIOTA_TYPE) {
          search.tablebase_type = value;
      }
#endif
#ifdef NALIMOV_TBS
      if (value == NALIMOV_TYPE) {
          search.tablebase_type = value;
      }
#endif
      if (search.tablebase_type == "") {
          cerr << "Invalid tablebase type: " << value << endl;
          search.tablebase_type = tmp;
      }
  }
#endif
#ifdef GAVIOTA_TBS
  else if (name == "search.gtb_cache_size") {
    setMemoryOption(search.gtb_cache_size,value);
  }
  else if (name == "search.gtb_path") {
    search.gtb_path = value;
  }
  else if (name == "search.gtb_scheme") {
    search.gtb_scheme = value;
  }
#endif
#ifdef NALIMOV_TBS
  else if (name == "search.nalimov_cache_size") {
    setMemoryOption(search.nalimov_cache_size,value);
  }
  else if (name == "search.nalimov_path") {
    search.nalimov_path = value;
  }
#endif
  else if (name == "search.strength") {
    set_strength_option(name,search.strength,value);
  }
  else if (name == "search.ncpus") {
    set_int_option(name,search.ncpus,value);
  }
  else
    cerr << "warning: unrecognized option name: " << name << endl;
}

int Options::init(const string& fileName) {
  
  ifstream infile(fileName.c_str());
  if (!infile.good()) {
     return -1;
  }
  string contents;
  while (infile.good() && !infile.eof()) {
      getline(infile,contents);
      string::const_iterator it = contents.begin();
      while (it != contents.end() && isspace(*it)) it++;
      if (it == contents.end()) break;
      if (*it == '#') continue;
      string name, value;
      while (it != contents.end() && *it != '=') {
          name += *it++;
      }
      if (it != contents.end()) it++;
      while (it != contents.end() && !isspace(*it)) {
          value += *it++;
      }
      if (name != "") {
         set_option(name,value);
      }
  }
  return 0;
}
